@@ -7,18 +7,23 @@
 # Eli Bendersky (eliben@gmail.com)
 # This code is in the public domain
 #-------------------------------------------------------------------------------
-import os, sys
-import re
+import argparse
 from difflib import SequenceMatcher
-from optparse import OptionParser
 import logging
+from multiprocessing import Pool
+import os
 import platform
-from utils import setup_syspath; setup_syspath()
+import re
+import sys
+import time
+
 from utils import run_exe, is_in_rootdir, dump_output_to_temp_files
 
+# Make it possible to run this file from the root dir of pyelftools without
+# installing pyelftools; useful for CI testing, etc.
+sys.path[0:0] = ['.']
 
 # Create a global logger object
-#
 testlog = logging.getLogger('run_tests')
 testlog.setLevel(logging.DEBUG)
 testlog.addHandler(logging.StreamHandler(sys.stdout))
@@ -26,9 +31,13 @@ testlog.addHandler(logging.StreamHandler(sys.stdout))
 # Set the path for calling readelf. We carry our own version of readelf around,
 # because binutils tend to change its output even between daily builds of the
 # same minor release and keeping track is a headache.
-READELF_PATH = 'test/external_tools/readelf'
-if not os.path.exists(READELF_PATH):
-    READELF_PATH = 'readelf'
+if platform.system() == "Darwin": # MacOS
+    READELF_PATH = 'greadelf'
+else:
+    READELF_PATH = 'test/external_tools/readelf'
+    if not os.path.exists(READELF_PATH):
+        READELF_PATH = 'readelf'
+
 
 def discover_testfiles(rootdir):
     """ Discover test files in the given directory. Yield them one by one.
@@ -39,17 +48,38 @@ def discover_testfiles(rootdir):
             yield os.path.join(rootdir, filename)
 
 
-def run_test_on_file(filename, verbose=False):
+def run_test_on_file(filename, verbose=False, opt=None):
     """ Runs a test on the given input filename. Return True if all test
         runs succeeded.
+        If opt is specified, rather that going over the whole
+        set of supported readelf options, the test will only
+        run for one option.
     """
     success = True
     testlog.info("Test file '%s'" % filename)
-    for option in [
-            '-e', '-d', '-s', '-r', '-x.text', '-p.shstrtab', '-V',
+    if opt is None:
+        options = [
+            '-e', '-d', '-s', '-n', '-r', '-x.text', '-p.shstrtab', '-V',
             '--debug-dump=info', '--debug-dump=decodedline',
-            '--debug-dump=frames', '--debug-dump=frames-interp']:
+            '--debug-dump=frames', '--debug-dump=frames-interp',
+            '--debug-dump=aranges', '--debug-dump=pubtypes',
+            '--debug-dump=pubnames', '--debug-dump=loc'
+            ]
+    else:
+        options = [opt]
+
+    for option in options:
         if verbose: testlog.info("..option='%s'" % option)
+
+        # TODO(zlobober): this is a dirty hack to make tests work for ELF core
+        # dump notes. Making it work properly requires a pretty deep
+        # investigation of how original readelf formats the output.
+        if "core" in filename and option == "-n":
+            if verbose:
+                testlog.warning("....will fail because corresponding part of readelf.py is not implemented yet")
+                testlog.info('.......................SKIPPED')
+            continue
+
         # stdouts will be a 2-element list: output of readelf and output
         # of scripts/readelf.py
         stdouts = []
@@ -57,13 +87,17 @@ def run_test_on_file(filename, verbose=False):
             args = [option, filename]
             if verbose: testlog.info("....executing: '%s %s'" % (
                 exe_path, ' '.join(args)))
+            t1 = time.time()
             rc, stdout = run_exe(exe_path, args)
+            if verbose: testlog.info("....elapsed: %s" % (time.time() - t1,))
             if rc != 0:
-                testlog.error("@@ aborting - '%s' returned '%s'" % (exe_path, rc))
+                testlog.error("@@ aborting - '%s %s' returned '%s'" % (exe_path, option, rc))
                 return False
             stdouts.append(stdout)
         if verbose: testlog.info('....comparing output...')
+        t1 = time.time()
         rc, errmsg = compare_output(*stdouts)
+        if verbose: testlog.info("....elapsed: %s" % (time.time() - t1,))
         if rc:
             if verbose: testlog.info('.......................SUCCESS')
         else:
@@ -92,31 +126,23 @@ def compare_output(s1, s2):
     """
     def prepare_lines(s):
         return [line for line in s.lower().splitlines() if line.strip() != '']
-    def filter_readelf_lines(lines):
-        filter_out = False
-        for line in lines:
-            if 'of the .eh_frame section' in line:
-                filter_out = True
-            elif 'of the .debug_frame section' in line:
-                filter_out = False
-            if not filter_out:
-                if not line.startswith('unknown: length'):
-                    yield line
 
     lines1 = prepare_lines(s1)
     lines2 = prepare_lines(s2)
 
-    lines1 = list(filter_readelf_lines(lines1))
-
-    flag_after_symtable = False
+    flag_in_debug_line_section = False
 
     if len(lines1) != len(lines2):
         return False, 'Number of lines different: %s vs %s' % (
                 len(lines1), len(lines2))
 
     for i in range(len(lines1)):
-        if 'symbol table' in lines1[i]:
-            flag_after_symtable = True
+        if lines1[i].endswith('debug_line section:'):
+            # .debug_line or .zdebug_line
+            flag_in_debug_line_section = True
+
+        # readelf spelling error for GNU property notes
+        lines1[i] = lines1[i].replace('procesor-specific type', 'processor-specific type')
 
         # Compare ignoring whitespace
         lines1_parts = lines1[i].split()
@@ -137,12 +163,32 @@ def compare_output(s1, s2):
             sm = SequenceMatcher()
             sm.set_seqs(lines1[i], lines2[i])
             changes = sm.get_opcodes()
-            if flag_after_symtable:
-                # Detect readelf's adding @ with lib and version after
-                # symbol name.
-                if (    len(changes) == 2 and changes[1][0] == 'delete' and
-                        lines1[i][changes[1][1]] == '@'):
+            if flag_in_debug_line_section:
+                # readelf outputs an additional "View" column: ignore it
+                if len(lines1_parts) >= 2 and lines1_parts[-2] == 'view':
                     ok = True
+                else:
+                    # Fast check special-cased for the only ELF we have which
+                    # has this information (dwarf_gnuops4.so.elf)
+                    ok = (    lines1_parts[-2:] == ['1', 'x']
+                          and lines2_parts[-1] == 'x')
+            elif '[...]' in lines1[i]:
+                # Special case truncations with ellipsis like these:
+                #     .note.gnu.bu[...]        redelf
+                #     .note.gnu.build-i        pyelftools
+                # Or more complex for symbols with versions, like these:
+                #     _unw[...]@gcc_3.0        readelf
+                #     _unwind_resume@gcc_3.0   pyelftools
+                for p1, p2 in zip(lines1_parts, lines2_parts):
+                    dots_start = p1.find('[...]')
+                    if dots_start != -1:
+                        break
+                ok = p1.endswith('[...]') and p1[:dots_start] == p2[:dots_start]
+                if not ok:
+                    dots_end = dots_start + 5
+                    if len(p1) > dots_end and p1[dots_end] == '@':
+                        ok = (    p1[:dots_start] == p2[:dots_start]
+                              and p1[p1.rfind('@'):] == p2[p2.rfind('@'):])
             elif 'at_const_value' in lines1[i]:
                 # On 32-bit machines, readelf doesn't correctly represent
                 # some boundary LEB128 numbers
@@ -153,11 +199,11 @@ def compare_output(s1, s2):
             elif 'os/abi' in lines1[i]:
                 if 'unix - gnu' in lines1[i] and 'unix - linux' in lines2[i]:
                     ok = True
-            elif (  'unknown at value' in lines1[i] and
-                    'dw_at_apple' in lines2[i]):
-                ok = True
+            elif len(lines1_parts) == 3 and lines1_parts[2] == 'nt_gnu_property_type_0':
+                # readelf does not seem to print a readable description for this
+                ok = lines1_parts == lines2_parts[:3]
             else:
-                for s in ('t (tls)', 'l (large)'):
+                for s in ('t (tls)', 'l (large)', 'd (mbind)'):
                     if s in lines1[i] or s in lines2[i]:
                         ok = True
                         break
@@ -173,38 +219,61 @@ def main():
         testlog.error('Error: Please run me from the root dir of pyelftools!')
         return 1
 
-    optparser = OptionParser(
-        usage='usage: %prog [options] [file] [file] ...',
+    argparser = argparse.ArgumentParser(
+        usage='usage: %(prog)s [options] [file] [file] ...',
         prog='run_readelf_tests.py')
-    optparser.add_option('-V', '--verbose',
-        action='store_true', dest='verbose',
-        help='Verbose output')
-    options, args = optparser.parse_args()
+    argparser.add_argument('files', nargs='*', help='files to run tests on')
+    argparser.add_argument(
+        '--parallel', action='store_true',
+        help='run tests in parallel; always runs all tests w/o verbose')
+    argparser.add_argument('-V', '--verbose',
+                           action='store_true', dest='verbose',
+                           help='verbose output')
+    argparser.add_argument(
+        '-k', '--keep-going',
+        action='store_true', dest='keep_going',
+        help="Run all tests, don't stop at the first failure")
+    argparser.add_argument('--opt',
+        action='store', dest='opt', metavar='<readelf-option>',
+        help= 'Limit the test one one readelf option.')
+    args = argparser.parse_args()
 
-    if options.verbose:
+    if args.parallel:
+        if args.verbose or args.keep_going == False:
+            print('WARNING: parallel mode disables verbosity and always keeps going')
+
+    if args.verbose:
         testlog.info('Running in verbose mode')
         testlog.info('Python executable = %s' % sys.executable)
         testlog.info('readelf path = %s' % READELF_PATH)
-        testlog.info('Given list of files: %s' % args)
+        testlog.info('Given list of files: %s' % args.files)
 
     # If file names are given as command-line arguments, only these files
     # are taken as inputs. Otherwise, autodiscovery is performed.
-    #
-    if len(args) > 0:
-        filenames = args
+    if len(args.files) > 0:
+        filenames = args.files
     else:
-        filenames = list(discover_testfiles('test/testfiles_for_readelf'))
+        filenames = sorted(discover_testfiles('test/testfiles_for_readelf'))
 
-    success = True
-    for filename in filenames:
-        if success:
-            success = success and run_test_on_file(
-                                    filename,
-                                    verbose=options.verbose)
+    if len(filenames) > 1 and args.parallel:
+        pool = Pool()
+        results = pool.map(run_test_on_file, filenames)
+        failures = results.count(False)
+    else:
+        failures = 0
+        for filename in filenames:
+            if not run_test_on_file(filename, args.verbose, args.opt):
+                failures += 1
+                if not args.keep_going:
+                    break
 
-    if success:
+    if failures == 0:
         testlog.info('\nConclusion: SUCCESS')
         return 0
+    elif args.keep_going:
+        testlog.info('\nConclusion: FAIL ({}/{})'.format(
+            failures, len(filenames)))
+        return 1
     else:
         testlog.info('\nConclusion: FAIL')
         return 1
@@ -212,4 +281,3 @@ def main():
 
 if __name__ == '__main__':
     sys.exit(main())
-
